@@ -23,16 +23,18 @@
 
 #include <aconf.h>
 
-#ifdef HAVE_FONT_EMBEDDING
-#  define GNOME_PRINT_UNSTABLE_API
-#endif
-
 #include <stdio.h>
 #include "Gfx.h"
 #include "GfxState.h"
 #include "GfxFont.h"
 #include "GPOutputDev.h"
 
+#ifdef HAVE_FONT_EMBEDDING
+#  include "gpdf-g-switch.h"
+#    include "gpdf-font-face.h"
+#  include "gpdf-g-switch.h"
+#endif
+
 #define noPDF_DEBUG
 #ifdef PDF_DEBUG
 #  define DBG(x) x
@@ -60,7 +62,7 @@ const struct {
   {"Courier-Bold",          "Nimbus Mono L Bold"},	      //"Courier Bold"},
   {"Courier-BoldOblique",   "Nimbus Mono L Bold Oblique"},  //"Courier Bold Oblique"},
   {"Symbol",                "Standard Symbols L Regular"},		      //"Symbol"}
-  {"ZapfDingbats",          "Dingbats "}      //"ITC Zapf Dingbats"}
+  {"ZapfDingbats",          "Dingbats Regular"}      //"ITC Zapf Dingbats"}
 };
 
 GPOFontMap::GPOFontMap() : embeddedFonts(NULL)
@@ -222,7 +224,7 @@ GnomeFontFace *GPOFontMap::getFontFaceEmbedded(GfxFont *font)
 
     DBG (g_message("Adding font face for %s", font_name));
     
-    gff = gnome_font_face_download ((const guchar *)font_name,
+    gff = gpdf_font_face_download ((const guchar *)font_name,
 				    (const guchar *)"",
 				    GNOME_FONT_REGULAR, FALSE,
 				    contents, length);
@@ -656,13 +658,15 @@ void GPOutputDev::drawString(GfxState *state, GString *s) /* FIXME */
     DBG (if (uLen > 1)
       g_warning("drawString: uLen > 1; p: %x", (gint)*p));
 
-    gnome_glyphlist_glyph(ggl,
-			  /* FIXME */
-			  gnome_font_lookup_default(gnome_font, u[0]));
-    gnome_glyphlist_rmoveto(ggl, dx, dy);
+    if (uLen == 1) {
+      gnome_glyphlist_glyph(ggl,
+			    /* FIXME */
+			    gnome_font_lookup_default(gnome_font, u[0]));
+      gnome_glyphlist_rmoveto(ggl, dx, dy);
 
-    text_pos_x += dx;
-    text_pos_y += dy;
+      text_pos_x += dx;
+      text_pos_y += dy;
+    } /* else FIXME */
 
     p += n;
     len -= n;
