@@ -17,10 +17,15 @@
 #include "gpdf-stock-icons.h"
 #include "gpdf-view.h"
 #include "gpdf-util.h"
-#include "page-control.h"
 #include "pdf-info-dict-util.h"
-#include "pdf-properties-display.h"
-#include <bonobo.h>
+#include "gpdf-g-switch.h"
+#  include "page-control.h"
+#  include "pdf-properties-display.h"
+#  include <bonobo.h>
+#  include <libgnomevfs/gnome-vfs-uri.h>
+#include "gpdf-g-switch.h"
+#include "gpdf-persist-stream.h"
+#include "gpdf-persist-file.h"
 
 #define noPDF_DEBUG
 #ifdef PDF_DEBUG
@@ -33,6 +38,9 @@ BEGIN_EXTERN_C
 
 struct _GPdfControlPrivate {
 	GPdfPersistStream *persist_stream;
+	GPdfPersistFile *persist_file;
+	PDFDoc *pdf_doc;
+	char *title;
 
 	GtkWidget *scrolled_window;
 	GtkWidget *gpdf_view;
@@ -40,17 +48,30 @@ struct _GPdfControlPrivate {
 	BonoboZoomable *zoomable;
 	gboolean has_zoomable_frame;
 
+	gboolean activated;
+	
 	BonoboControl *bonobo_page_control;
 
 	guint idle_id;
 };
 
 #define PARENT_TYPE BONOBO_TYPE_CONTROL
-static BonoboControlClass *parent_class;
+BONOBO_CLASS_BOILERPLATE (GPdfControl, gpdf_control,
+			  BonoboControl, PARENT_TYPE);
 
 #define GPDF_IS_NON_NULL_CONTROL(obj) \
 (((obj) != NULL) && (GPDF_IS_CONTROL ((obj))))
 
+enum {
+	PROP_0,
+	PROP_PERSIST_STREAM,
+	PROP_PERSIST_FILE
+};
+
+enum {
+	PBPROP_TITLE
+};
+
 
 static void
 gpdf_control_properties_dialog_response_cb (GtkDialog *dialog, gint response_id,
@@ -64,7 +85,6 @@ static void
 verb_FileProperties_cb (BonoboUIComponent *uic, gpointer user_data,
 			const char *cname)
 {
-	GPdfPersistStream *persist_stream;
 	GtkWidget *dialog;
 	GtkWidget *props;
 
@@ -84,11 +104,9 @@ verb_FileProperties_cb (BonoboUIComponent *uic, gpointer user_data,
 	bonobo_control_set_transient_for (BONOBO_CONTROL (user_data),
 					  GTK_WINDOW (dialog), NULL);
 
-	persist_stream = GPDF_CONTROL (user_data)->priv->persist_stream;
 	props = GTK_WIDGET (g_object_new (GPDF_TYPE_PROPERTIES_DISPLAY, NULL));
 	pdf_doc_process_properties (
-		gpdf_persist_stream_get_pdf_doc (persist_stream),
-		gpdf_persist_stream_get_length (persist_stream),
+		GPDF_CONTROL (user_data)->priv->pdf_doc,
 		G_OBJECT (props));
 
 	gtk_widget_show (props);
@@ -97,6 +115,26 @@ verb_FileProperties_cb (BonoboUIComponent *uic, gpointer user_data,
 }
 
 
+static void
+verb_GoHistoryPrev_cb (BonoboUIComponent *uic, gpointer user_data,
+		       const char *cname)
+{
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
+
+	gpdf_view_back_history (
+		GPDF_VIEW (GPDF_CONTROL (user_data)->priv->gpdf_view));
+}
+
+static void
+verb_GoHistoryNext_cb (BonoboUIComponent *uic, gpointer user_data,
+		       const char *cname)
+{
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
+
+	gpdf_view_forward_history (
+		GPDF_VIEW (GPDF_CONTROL (user_data)->priv->gpdf_view));
+}
+
 static void
 verb_GoPagePrev_cb (BonoboUIComponent *uic, gpointer user_data, const char*cname)
 {
@@ -172,17 +210,52 @@ gc_enable_go_forward_verbs (GPdfControl *control, gboolean sensitive)
 	ui_component_set_sensitive (ui, "/commands/GoPageLast", sensitive);
 }
 
+static void
+gc_enable_history_back_verbs (GPdfControl *control, gboolean sensitive)
+{
+	BonoboUIComponent *ui;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (control));
+
+	ui = bonobo_control_get_ui_component (BONOBO_CONTROL (control));
+	if (ui == NULL)
+		return;
+
+	ui_component_set_sensitive (ui, "/commands/GoHistoryPrev", sensitive);
+}
+
+static void
+gc_enable_history_forward_verbs (GPdfControl *control, gboolean sensitive)
+{
+	BonoboUIComponent *ui;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (control));
+
+	ui = bonobo_control_get_ui_component (BONOBO_CONTROL (control));
+	if (ui == NULL)
+		return;
+
+	ui_component_set_sensitive (ui, "/commands/GoHistoryNext", sensitive);
+}
+
 static void
 gc_page_changed_enable_page_buttons (GPdfView *gpdf_view, gint page,
 				     GPdfControl *control)
 {
 	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (control));
 
+	if (!control->priv->activated)
+		return;
+
 	gpdf_view = GPDF_VIEW (control->priv->gpdf_view);
 
 	gc_enable_go_back_verbs (control, !gpdf_view_is_first_page (gpdf_view));
 	gc_enable_go_forward_verbs (control,
 				    !gpdf_view_is_last_page (gpdf_view));
+	gc_enable_history_back_verbs (control,
+				      !gpdf_view_is_first_history (gpdf_view));
+	gc_enable_history_forward_verbs (control,
+				    !gpdf_view_is_last_history (gpdf_view));
 }
 
 
@@ -451,6 +524,9 @@ gc_set_zoom_items_visibility (GPdfControl *control, gboolean visible)
 static BonoboUIVerb gc_verbs [] = {
 	BONOBO_UI_VERB ("FileProperties", verb_FileProperties_cb),
 
+	BONOBO_UI_VERB ("GoHistoryPrev",     verb_GoHistoryPrev_cb),
+	BONOBO_UI_VERB ("GoHistoryNext",     verb_GoHistoryNext_cb),
+
 	BONOBO_UI_VERB ("GoPagePrev",     verb_GoPagePrev_cb),
 	BONOBO_UI_VERB ("GoPageNext",     verb_GoPageNext_cb),
 	BONOBO_UI_VERB ("GoPageFirst",    verb_GoPageFirst_cb),
@@ -593,10 +669,96 @@ gpdf_control_activate (BonoboControl *control, gboolean activate)
 		gc_unset_ui_container (gpdf_control);
 	}
 
+	gpdf_control->priv->activated = activate;
+
+	/* FIXME Ugly fix to setup controls enable status */
+	gc_page_changed_enable_page_buttons (
+		GPDF_VIEW (gpdf_control->priv->gpdf_view),
+		1,
+		gpdf_control);
+
 	BONOBO_CALL_PARENT (
 		BONOBO_CONTROL_CLASS, activate, (control, activate));
 }
 
+static void
+gpdf_control_pdf_doc_changed (GPdfControl *gpdf_control)
+{
+	GPdfControlPrivate *priv;
+	GPdfView *gpdf_view;
+	GPdfPageControl *page_control;
+
+	priv = gpdf_control->priv;
+
+	gpdf_view = GPDF_VIEW (priv->gpdf_view);
+	page_control = GPDF_PAGE_CONTROL (
+		bonobo_control_get_widget (priv->bonobo_page_control));
+	if (priv->pdf_doc) {
+		gpdf_view_set_pdf_doc (gpdf_view, priv->pdf_doc);
+		gpdf_page_control_set_total_pages (
+			page_control, priv->pdf_doc->getNumPages ());
+	} else {
+		gpdf_page_control_set_total_pages (page_control, -1);
+	}
+
+}
+
+static void
+persist_stream_set_pdf_cb (GPdfPersistStream *persist_stream,
+			   gpointer user_data)
+{
+	GPdfControl *control;
+
+	g_return_if_fail (persist_stream != NULL);
+	g_return_if_fail (GPDF_IS_PERSIST_STREAM (persist_stream));
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
+
+	control = GPDF_CONTROL (user_data);
+	control->priv->pdf_doc =
+		gpdf_persist_stream_get_pdf_doc (persist_stream);
+
+	if (control->priv->title)
+		g_free (control->priv->title);
+
+	control->priv->title = g_strdup (_("PDF Document"));
+
+	gpdf_control_pdf_doc_changed (control);
+}
+
+static void
+persist_file_set_pdf_cb (GPdfPersistFile *persist_file, gpointer user_data)
+{
+	GPdfControl *control;
+	const char *uri;
+	GnomeVFSURI *vfs_uri;
+
+	g_return_if_fail (persist_file != NULL);
+	g_return_if_fail (GPDF_IS_PERSIST_FILE (persist_file));
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
+
+	control = GPDF_CONTROL (user_data);
+	control->priv->pdf_doc = gpdf_persist_file_get_pdf_doc (persist_file);
+
+	if (control->priv->title)
+		g_free (control->priv->title);
+
+	uri = gpdf_persist_file_get_current_uri (persist_file);
+
+	vfs_uri = uri ? gnome_vfs_uri_new (uri) : NULL;
+
+	if (vfs_uri) {
+		control->priv->title = gnome_vfs_uri_is_local (vfs_uri) ?
+			g_strdup (gnome_vfs_uri_get_path (vfs_uri)) :
+			gnome_vfs_uri_to_string (vfs_uri,
+						 GNOME_VFS_URI_HIDE_NONE);
+                gnome_vfs_uri_unref (vfs_uri);
+	} else {
+		control->priv->title = g_strdup (_("PDF Document"));
+	}
+
+	gpdf_control_pdf_doc_changed (control);
+}
+
 static void
 gpdf_control_destroy (BonoboObject *object)
 {
@@ -613,79 +775,108 @@ gpdf_control_destroy (BonoboObject *object)
 		priv->bonobo_page_control = NULL;
 	}
 
+	if (priv->title) {
+		g_free (priv->title);
+		priv->title = NULL;
+	}
+
 	BONOBO_CALL_PARENT (BONOBO_OBJECT_CLASS, destroy, (object));
 }
 
 static void
-gpdf_control_finalize (GObject *object)
+gpdf_control_set_property (GObject *object, guint param_id,
+			   const GValue *value, GParamSpec *pspec)
 {
+	GPdfControl *gpdf_control;
+
 	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (object));
 
-	g_free ((GPDF_CONTROL (object))->priv);
+	gpdf_control = GPDF_CONTROL (object);
 
-	BONOBO_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
-}
+	switch (param_id) {
+	case PROP_PERSIST_STREAM: {
+		GPdfPersistStream *pstream;
+		
+		pstream = GPDF_PERSIST_STREAM (g_value_get_object (value));
+		
+		gpdf_control->priv->persist_stream = pstream;
 
-static void
-gpdf_control_class_init (GPdfControlClass *klass)
-{
-	parent_class = BONOBO_CONTROL_CLASS (gtk_type_class (PARENT_TYPE));
+		bonobo_object_ref (BONOBO_OBJECT (pstream));
+		bonobo_object_add_interface (BONOBO_OBJECT (gpdf_control),
+					     BONOBO_OBJECT (pstream));
+		break;
+	}
+	case PROP_PERSIST_FILE: {
+		GPdfPersistFile *pfile;
 
-	G_OBJECT_CLASS (klass)->finalize = gpdf_control_finalize;
-	BONOBO_OBJECT_CLASS (klass)->destroy = gpdf_control_destroy;
-	BONOBO_CONTROL_CLASS (klass)->activate = gpdf_control_activate;
-}
+		pfile = GPDF_PERSIST_FILE (g_value_get_object (value));
 
-static void
-gpdf_control_init (GPdfControl *control)
-{
-	control->priv = g_new0 (GPdfControlPrivate, 1);
-}
+		gpdf_control->priv->persist_file = pfile;
 
-static void
-gc_set_pdf_cb (GPdfPersistStream *persist_stream, gpointer user_data)
-{
-	PDFDoc *pdf_doc = NULL;
-	GPdfControlPrivate *priv;
-	GPdfView *gpdf_view;
-	GPdfPageControl *page_control;
-
-	g_return_if_fail (persist_stream != NULL);
-	g_return_if_fail (GPDF_IS_PERSIST_STREAM (persist_stream));
-	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
-
-	pdf_doc = gpdf_persist_stream_get_pdf_doc (persist_stream);
-
-	priv = GPDF_CONTROL (user_data)->priv;
-	gpdf_view = GPDF_VIEW (priv->gpdf_view);
-	page_control = GPDF_PAGE_CONTROL (
-		bonobo_control_get_widget (priv->bonobo_page_control));
-	if (pdf_doc) {
-		gpdf_view_set_pdf_doc (gpdf_view, pdf_doc);
-		gpdf_page_control_set_total_pages (page_control,
-						   pdf_doc->getNumPages ());
-	} else {
-		gpdf_page_control_set_total_pages (page_control, -1);
+		bonobo_object_ref (BONOBO_OBJECT (pfile));
+		bonobo_object_add_interface (BONOBO_OBJECT (gpdf_control),
+					     BONOBO_OBJECT (pfile));
+		break;
+	}
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+		break;
 	}
 }
 
 static void
-gc_setup_persist (GPdfControl *control, GPdfPersistStream *persist_stream)
+gpdf_control_setup_persist (GPdfControl *control)
 {
-	control->priv->persist_stream = persist_stream;
-
-	g_signal_connect (G_OBJECT (persist_stream),
+	g_signal_connect (G_OBJECT (control->priv->persist_stream),
 			  "set_pdf",
-			  G_CALLBACK (gc_set_pdf_cb),
+			  G_CALLBACK (persist_stream_set_pdf_cb),
+			  control);
+	g_signal_connect (G_OBJECT (control->priv->persist_file),
+			  "set_pdf",
+			  G_CALLBACK (persist_file_set_pdf_cb),
 			  control);
-
-	bonobo_object_ref (BONOBO_OBJECT (persist_stream));
-	bonobo_object_add_interface (BONOBO_OBJECT (control),
-				     BONOBO_OBJECT (persist_stream));
 }
 
 static void
-gc_setup_view_widgets (GPdfControl *control, GPdfPersistStream *persist_stream)
+property_bag_get_prop (BonoboPropertyBag *bag, BonoboArg *arg, guint arg_id,
+		       CORBA_Environment *ev, gpointer user_data)
+{
+	GPdfControl *control;
+
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (user_data));
+
+	control = GPDF_CONTROL (user_data);
+
+	switch (arg_id) {
+	case PBPROP_TITLE:
+		BONOBO_ARG_SET_STRING (
+			arg, 
+			CORBA_string_dup (control->priv->title));
+		break;
+	default:
+		break;
+	}		
+}
+
+static void
+gpdf_control_setup_property_bag (GPdfControl *gpdf_control)
+{
+	BonoboPropertyBag *property_bag;
+
+	property_bag = bonobo_property_bag_new (property_bag_get_prop, NULL,
+						gpdf_control);
+	bonobo_property_bag_add (property_bag, "title",
+				 PBPROP_TITLE, TC_CORBA_string, NULL,
+				 "Document title",
+				 BONOBO_PROPERTY_READABLE);
+	bonobo_control_set_properties (BONOBO_CONTROL (gpdf_control),
+				       BONOBO_OBJREF (property_bag),
+				       NULL);
+	bonobo_object_unref (BONOBO_OBJECT (property_bag));
+}
+
+static void
+gpdf_control_setup_view_widgets (GPdfControl *control)
 {
 	GtkWidget *sw;
 	GtkWidget *gpdf_view;
@@ -709,7 +900,7 @@ gc_setup_view_widgets (GPdfControl *control, GPdfPersistStream *persist_stream)
 }
 
 static void
-gc_setup_zoomable (GPdfControl *control)
+gpdf_control_setup_zoomable (GPdfControl *control)
 {
 	BonoboZoomable *zoomable = bonobo_zoomable_new ();
 
@@ -753,7 +944,7 @@ gc_setup_zoomable (GPdfControl *control)
 }
 
 static void
-gc_setup_page_control (GPdfControl *control)
+gpdf_control_setup_page_control (GPdfControl *control)
 {
 	GtkWidget *page_control;
 	BonoboControl *bonobo_page_control;
@@ -774,40 +965,82 @@ gc_setup_page_control (GPdfControl *control)
 	control->priv->bonobo_page_control = bonobo_page_control;
 }
 
-GPdfControl *
-gpdf_control_construct (GPdfControl *control,
-			GPdfPersistStream *persist_stream)
+static GObject *
+gpdf_control_constructor (GType type,
+			  guint n_construct_properties,
+			  GObjectConstructParam *construct_params)
 {
-	g_return_val_if_fail (GPDF_IS_NON_NULL_CONTROL (control), NULL);
-	g_return_val_if_fail (persist_stream != NULL, NULL);
-	g_return_val_if_fail (GPDF_IS_PERSIST_STREAM (persist_stream), NULL);
+	GObject *object;
+	GPdfControl *control;
+	GPdfControlPrivate *priv;
+	guint i;
 
-	gc_setup_persist (control, persist_stream);
-	gc_setup_view_widgets (control, persist_stream);
-	gc_setup_zoomable (control);
-	gc_setup_page_control (control);
-	bonobo_control_construct (BONOBO_CONTROL (control),
+	object = G_OBJECT_CLASS (parent_class)->constructor (
+		type, n_construct_properties, construct_params);
+
+	control = GPDF_CONTROL (object);
+	priv = control->priv;
+
+	g_return_val_if_fail (priv->persist_stream != NULL, NULL);
+	g_return_val_if_fail (priv->persist_file != NULL, NULL);
+	gpdf_control_setup_persist (control);
+	gpdf_control_setup_view_widgets (control);
+	gpdf_control_setup_zoomable (control);
+	gpdf_control_setup_page_control (control);
+
+	bonobo_control_construct (BONOBO_CONTROL (object),
 				  control->priv->scrolled_window);
 
-	return control;
+	control->priv->title = g_strdup (_("PDF Document"));
+	gpdf_control_setup_property_bag (control);
+
+	return object;
 }
 
-GPdfControl *
-gpdf_control_new (GPdfPersistStream *gpdf_persist_stream)
+static void
+gpdf_control_finalize (GObject *object)
 {
-	GPdfControl *control;
+	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (object));
 
-	g_return_val_if_fail (gpdf_persist_stream != NULL, NULL);
-	g_return_val_if_fail (GPDF_IS_PERSIST_STREAM (gpdf_persist_stream),
-			      NULL);
+	g_free ((GPDF_CONTROL (object))->priv);
 
-	DBG (g_message ("Creating GPdfControl"));
-
-	control = GPDF_CONTROL (g_object_new (GPDF_TYPE_CONTROL, NULL));
-
-	return gpdf_control_construct (control, gpdf_persist_stream);
+	BONOBO_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
 }
 
-BONOBO_TYPE_FUNC (GPdfControl, PARENT_TYPE, gpdf_control)
+static void
+gpdf_control_class_init (GPdfControlClass *klass)
+{
+	GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+	object_class->constructor = gpdf_control_constructor;
+	object_class->finalize = gpdf_control_finalize;
+	object_class->set_property = gpdf_control_set_property;
+
+	BONOBO_OBJECT_CLASS (klass)->destroy = gpdf_control_destroy;
+	BONOBO_CONTROL_CLASS (klass)->activate = gpdf_control_activate;
+
+	g_object_class_install_property (
+		object_class, PROP_PERSIST_STREAM,
+		g_param_spec_object ("persist_stream",
+				     "PersistStream",
+				     "PersistStream",
+				     GPDF_TYPE_PERSIST_STREAM,
+				     (GParamFlags)(G_PARAM_WRITABLE |
+						   G_PARAM_CONSTRUCT_ONLY)));
+	g_object_class_install_property (
+		object_class, PROP_PERSIST_FILE,
+		g_param_spec_object ("persist_file",
+				     "PersistFile",
+				     "PersistFile",
+				     GPDF_TYPE_PERSIST_FILE,
+				     (GParamFlags)(G_PARAM_WRITABLE |
+						   G_PARAM_CONSTRUCT_ONLY)));
+}
+
+static void
+gpdf_control_instance_init (GPdfControl *control)
+{
+	control->priv = g_new0 (GPdfControlPrivate, 1);
+}
 
 END_EXTERN_C
