@@ -13,10 +13,12 @@
 #  undef GString
 #  undef GList
 #  undef GDir
+#  undef GMutex
 #  undef GPDF_GOO
 #else
 #  define GString G_String
 #  define GList   G_List
 #  define GDir    G_Dir
+#  define GMutex  G_Mutex
 #  define GPDF_GOO
 #endif
