@@ -31,10 +31,13 @@
 #  include <libgnome/gnome-macros.h>
 #  include <libgnomeprint/gnome-print.h>
 #  include <libgnomeprintui/gnome-print-preview.h>
+#  include <ggv-document.h>
 #  include "gtkgesture.h"
 #include "gpdf-g-switch.h"
 #include "GPOutputDev.h"
 #include "PDFDoc.h"
+#include "Link.h"
+#include "gpdf-links-canvas-layer.h"
 
 BEGIN_EXTERN_C
 
@@ -43,6 +46,8 @@ struct _GPdfViewPrivate {
 	GPOutputDev *output_dev;
 	PDFDoc *pdf_doc;
 
+	GPdfLinksCanvasLayer *links_layer;
+
 	GnomeCanvasItem *page_background;
 	GnomeCanvasItem *page_shadow;
 
@@ -54,18 +59,23 @@ struct _GPdfViewPrivate {
 	guint dragging : 1;
 	gint drag_anchor_x, drag_anchor_y;
 	gint drag_offset_x, drag_offset_y;
+
+	G_List *forward_history; 
+	G_List *backward_history; 
 };
 
 #define PARENT_TYPE GNOME_TYPE_CANVAS
-GPDF_CLASS_BOILERPLATE (GPdfView, gpdf_view, GnomeCanvas, PARENT_TYPE);
+static GnomeCanvasClass *parent_class = NULL;
 
 #define GPDF_IS_NON_NULL_VIEW(obj) (((obj) != NULL) && (GPDF_IS_VIEW ((obj))))
 #define GPDF_IS_VIEW_WITH_DOC(obj) \
-(GPDF_IS_NON_NULL_VIEW (obj) && (obj)->priv->pdf_doc)
+(GPDF_IS_NON_NULL_VIEW (obj) && ((GPdfView *)(obj))->priv->pdf_doc)
 
 enum {
 	ZOOM_CHANGED_SIGNAL,
 	PAGE_CHANGED_SIGNAL,
+	CLOSE_REQUESTED_SIGNAL,
+	QUIT_REQUESTED_SIGNAL,
 	LAST_SIGNAL
 };
 
@@ -107,26 +117,192 @@ gpdf_view_page_background_resize (GPdfView *gpdf_view,
 }
 
 static void
-setup_upside_down_transform (gdouble *transform, gdouble height)
+setup_page_transform (gdouble *transform, 
+		      gdouble x1, gdouble y1, gdouble x2, gdouble y2,
+		      gint angle)
 {
-	transform [0] = 1.0;
-	transform [1] = 0.0;
-	transform [2] = 0.0;
-	transform [3] = -1.0;
-	transform [4] = 0.0;
-	transform [5] = height;
+	double width;
+	double height;
+	double rotate [6];
+	double translate [6];
+
+	width = x2 - x1;
+	height = y2 - y1;
+
+	art_affine_identity (transform);
+	art_affine_flip (transform, transform, FALSE, TRUE);
+	art_affine_rotate (rotate, angle);
+	art_affine_multiply (transform, transform, rotate);
+
+	if (angle == 90) {
+		transform [4] = 0.0;
+		transform [5] = 0.0;
+	} else if (angle == 180) {
+		transform [4] = width;
+		transform [5] = 0.0;
+	} else if (angle == 270) {
+		transform [4] = height;
+		transform [5] = width;
+	} else /* if (angle == 0) */ {
+		transform [4] = 0.0;
+		transform [5] = height;
+	}
+
+	art_affine_translate (translate, -x1, -y1);
+	art_affine_multiply (transform, translate, transform);
 }	
 
+static gint
+canonical_multiple_of_90 (gint n)
+{
+	while (n >= 360) {
+		n -= 360;
+	}
+	while (n < 0) {
+		n += 360;
+	}
+
+	return 90 * (gint)((n / 90.0) + .5);
+}
+
+static void
+gpdf_free_history_node (gpointer node, gpointer user_data)
+{
+	g_free (node); 
+}
+
+static void 
+gpdf_view_history_stack_page (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv = gpdf_view->priv;
+	GPdfViewHistory *hnode; 
+
+	/* First delete forward history if one exists */
+	if (priv->forward_history) {
+		g_list_foreach (priv->forward_history,
+				gpdf_free_history_node,
+				NULL); 
+		g_list_free (priv->forward_history);
+		priv->forward_history = (G_List *)NULL;
+	}
+
+	/* Prepend history node */
+	hnode = (GPdfViewHistory *)g_malloc (sizeof(GPdfViewHistory));
+	hnode->zoom = priv->zoom;
+	hnode->page = priv->current_page;
+	priv->backward_history = g_list_prepend (priv->backward_history, (gpointer)hnode); 
+}
+
+static void
+gpdf_view_link_action_goto (GPdfView *gpdf_view, LinkGoTo *link_goto)
+{
+	PDFDoc *pdf_doc;
+	LinkDest *dest;
+	GString *named_dest;
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	g_return_if_fail (dynamic_cast <LinkGoTo *> (link_goto) != NULL);
+
+	pdf_doc = gpdf_view->priv->pdf_doc;
+	dest = link_goto->getDest ();
+	named_dest = link_goto->getNamedDest ();
+
+	if (dest != NULL) {
+		dest = dest->copy();
+	} else if (named_dest != NULL) {
+		named_dest = named_dest->copy();
+		dest = pdf_doc->findDest (named_dest);
+		delete named_dest;
+	}
+
+	if (dest != NULL) {
+		Ref page_ref;
+		int page;
+
+		if (dest->isPageRef ()) {
+			page_ref = dest->getPageRef ();
+			page = pdf_doc->findPage (page_ref.num, page_ref.gen);
+		} else {
+			page = dest->getPageNum ();
+		}
+
+		delete dest;
+
+		if (page != gpdf_view->priv->requested_page)
+			gpdf_view_goto_page (gpdf_view, page);
+	}	
+}
+
+static void
+gpdf_view_link_action_named (GPdfView *gpdf_view, LinkNamed *link_named)
+{
+	const char *action_name;
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	g_return_if_fail (dynamic_cast <LinkNamed *> (link_named) != NULL);
+
+	action_name = link_named->getName ()->getCString ();
+	
+	if (strcmp (action_name, "NextPage") == 0) {
+		gpdf_view_page_next (gpdf_view);
+	} else if (strcmp (action_name, "PrevPage") == 0) {
+		gpdf_view_page_prev (gpdf_view);
+	} else if (strcmp (action_name, "FirstPage") == 0) {
+		gpdf_view_page_first (gpdf_view);
+	} else if (strcmp (action_name, "LastPage") == 0) {
+		gpdf_view_page_last (gpdf_view);
+	} else if (strcmp (action_name, "Close") == 0) {
+		g_signal_emit (
+			G_OBJECT (gpdf_view), CLOSE_REQUESTED_SIGNAL, 0);
+	} else if (strcmp (action_name, "Quit") == 0) {
+		g_signal_emit (G_OBJECT (gpdf_view), QUIT_REQUESTED_SIGNAL, 0);
+	} else {
+		g_message ("Unimplemented named link action: %s", action_name);
+	}
+}
+
+static void
+link_clicked_cb (GPdfLinksCanvasLayer *links_layer, Link *link,
+		 gpointer user_data)
+{
+	GPdfView *gpdf_view;
+	LinkAction *link_action;
+
+	g_return_if_fail (GPDF_IS_VIEW (user_data));
+	g_return_if_fail (dynamic_cast <Link *> (link) != NULL);
+
+	gpdf_view = GPDF_VIEW (user_data);
+	link_action = link->getAction ();
+
+	switch (link_action->getKind ()) {
+	case actionGoTo:
+		gpdf_view_link_action_goto (
+			gpdf_view, dynamic_cast <LinkGoTo *> (link_action));
+		break;
+	case actionNamed:
+		gpdf_view_link_action_named (
+			gpdf_view, dynamic_cast <LinkNamed *> (link_action));
+		break;
+	default:
+		g_message ("Click on link: unimplemented Action");
+		break;
+	}
+}
+
 static gboolean
 gpdf_view_render_page (GPdfView *gpdf_view)
 {
 	GPdfViewPrivate *priv;
 	GnomePrintConfig *config;
 	PDFDoc *pdf_doc;
+	Page *pdf_page;
+	const PDFRectangle *rect;
 	gint page;
 	gint page_rotate;
+	gdouble page_width_rot, page_height_rot;
 	ArtDRect region;
 	gdouble page_transform [6];
+	Object obj;
 
 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), FALSE);
 
@@ -143,36 +319,34 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 	}
 
 	pdf_doc = priv->pdf_doc;
+	pdf_page = pdf_doc->getCatalog ()->getPage (page);
+	rect = pdf_page->getBox ();
 
-	config = gnome_print_config_default ();
+	page_rotate = canonical_multiple_of_90 (pdf_doc->getPageRotate (page));
 
-	gnome_print_config_set_length (
-		config,
-		(const guchar *)GNOME_PRINT_KEY_LAYOUT_WIDTH,
-		pdf_doc->getPageWidth (page),
-		GNOME_PRINT_PS_UNIT);
+	setup_page_transform (page_transform,
+			      rect->x1, rect->y1, rect->x2, rect->y2,
+			      page_rotate);
 
-	gnome_print_config_set_length (
-		config,
-		(const guchar *)GNOME_PRINT_KEY_LAYOUT_HEIGHT,
-		pdf_doc->getPageHeight (page),
-		GNOME_PRINT_PS_UNIT);
-
-	setup_upside_down_transform (page_transform,
-				     pdf_doc->getPageHeight (page));
+	if (page_rotate == 90 || page_rotate == 270) {
+		page_width_rot = pdf_doc->getPageHeight (page);
+		page_height_rot = pdf_doc->getPageWidth (page);
+	} else /* if (page_rotate == 0 || page_rotate == 180) */ {
+		page_width_rot = pdf_doc->getPageWidth (page);
+		page_height_rot = pdf_doc->getPageHeight (page);
+	}
 
 	/* Set scrolling region */
 	region.x0 = region.y0 = 0.0 - PAGE_PAD;
-	region.x1 = pdf_doc->getPageWidth (page) + PAGE_PAD;
-	region.y1 = pdf_doc->getPageHeight (page) + PAGE_PAD;
+	region.x1 = page_width_rot + PAGE_PAD;
+	region.y1 = page_height_rot + PAGE_PAD;
 
-	gpdf_view_page_background_resize (gpdf_view, 
-					  pdf_doc->getPageWidth (page),
-					  pdf_doc->getPageHeight (page));
+	gpdf_view_page_background_resize (gpdf_view,
+					  page_width_rot, page_height_rot);
 
+	config = gnome_print_config_default ();
 	priv->print_context = gnome_print_preview_new_full (
 		config, GNOME_CANVAS (gpdf_view), page_transform, &region);
-
 	gnome_print_config_unref (config);
 
 	priv->output_dev->setPrintContext (priv->print_context);
@@ -184,6 +358,30 @@ gpdf_view_render_page (GPdfView *gpdf_view)
 
 	priv->current_page = page;
 
+	if (priv->links_layer) {
+		g_object_unref (G_OBJECT (priv->links_layer));
+		gtk_object_destroy (GTK_OBJECT (priv->links_layer));
+		priv->links_layer = NULL;
+	}
+
+	priv->links_layer = GPDF_LINKS_CANVAS_LAYER (gnome_canvas_item_new (
+		gnome_canvas_root (GNOME_CANVAS (gpdf_view)),
+		GPDF_TYPE_LINKS_CANVAS_LAYER,
+		"links",
+		new Links (pdf_page->getAnnots (&obj), 
+			   pdf_doc->getCatalog ()->getBaseURI ()),
+		NULL));
+	g_object_ref (G_OBJECT (priv->links_layer));
+	gtk_object_sink (GTK_OBJECT (priv->links_layer));
+
+	obj.free ();
+
+	gnome_canvas_item_affine_relative (
+		GNOME_CANVAS_ITEM (priv->links_layer), page_transform);
+
+	g_signal_connect (G_OBJECT (priv->links_layer), "link_clicked",
+			  G_CALLBACK (link_clicked_cb), gpdf_view);
+
 	return FALSE;
 }
 
@@ -195,8 +393,8 @@ gpdf_view_emit_page_changed (GPdfView *gpdf_view, gint page)
 		       0, page);
 }
 
-void
-gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
+static void
+gpdf_view_goto_page_no_history (GPdfView *gpdf_view, gint page)
 {
 	GPdfViewPrivate *priv;
 	gint last_page;
@@ -209,6 +407,15 @@ gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
 	g_idle_add ((GSourceFunc)gpdf_view_render_page, gpdf_view);
 }
 
+void
+gpdf_view_goto_page (GPdfView *gpdf_view, gint page)
+{
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+
+	gpdf_view_history_stack_page (gpdf_view); 
+	gpdf_view_goto_page_no_history (gpdf_view, page);
+}
+
 void
 gpdf_view_page_prev (GPdfView *gpdf_view)
 {
@@ -262,6 +469,115 @@ gpdf_view_is_last_page (GPdfView *gpdf_view)
 	return gpdf_view->priv->requested_page == last_page;
 }
 
+void
+gpdf_view_back_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+	GPdfViewHistory *hnode;
+	int target_page = 0; 
+	double target_zoom = (double)0.0; 
+	
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	priv = gpdf_view->priv;
+
+	g_return_if_fail (priv->backward_history != (G_List *)NULL);
+	g_return_if_fail (priv->backward_history->data != (gpointer)NULL); 
+
+	hnode = (GPdfViewHistory *)priv->backward_history->data;
+	target_page = hnode->page; 
+	target_zoom = hnode->zoom; 
+	priv->backward_history = g_list_remove (priv->backward_history,
+						(gpointer)hnode);
+	if (!g_list_length (priv->backward_history)) {
+		g_list_free (priv->backward_history);
+		priv->backward_history = (G_List *)NULL;
+	}
+	hnode->page = priv->current_page; 
+	hnode->zoom = priv->zoom; 
+	priv->forward_history = g_list_prepend (priv->forward_history, hnode); 
+
+	gpdf_view_goto_page_no_history (gpdf_view, target_page);
+	/* TODO: change zoom level if different */
+}
+
+void
+gpdf_view_forward_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+	GPdfViewHistory *hnode; 
+	int target_page = 0; 
+	double target_zoom = (double)0.0; 
+
+	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
+	priv = gpdf_view->priv;
+
+	g_return_if_fail (priv->forward_history != (G_List *)NULL); 
+	g_return_if_fail (priv->forward_history->data != (gpointer)NULL); 
+	hnode = (GPdfViewHistory *)priv->forward_history->data;
+	target_page = hnode->page; 
+	target_zoom = hnode->zoom; 
+	priv->forward_history = g_list_remove (priv->forward_history,
+					       (gpointer)hnode);
+	if (!g_list_length (priv->forward_history)) {
+		g_list_free (priv->forward_history);
+		priv->forward_history = (G_List *)NULL;
+	}
+	hnode->page = priv->current_page; 
+	hnode->zoom = priv->zoom; 
+	priv->backward_history = g_list_prepend (priv->backward_history, hnode); 
+
+	gpdf_view_goto_page_no_history (gpdf_view, target_page);
+	/* TODO: change zoom level if different */
+}
+
+gboolean
+gpdf_view_is_first_history(GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), TRUE);
+	priv = gpdf_view->priv;
+
+	return (priv->backward_history == (G_List *)NULL); 
+}
+
+gboolean
+gpdf_view_is_last_history (GPdfView *gpdf_view)
+{
+	GPdfViewPrivate *priv;
+
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view), TRUE);
+	priv = gpdf_view->priv;
+
+	return (priv->forward_history == (G_List *)NULL); 
+}
+
+static int
+gpdf_view_get_page_count (GgvDocument *document)
+{
+ 	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), -1);
+
+	return GPDF_VIEW (document)->priv->pdf_doc->getNumPages ();
+}
+
+static char **
+gpdf_view_get_page_names (GgvDocument *document)
+{
+	int page_count;
+	int i;
+	char **result;
+	
+	g_return_val_if_fail (GPDF_IS_VIEW_WITH_DOC (document), NULL);
+
+	page_count = gpdf_view_get_page_count (document);
+
+	result = g_new0 (char *, page_count + 1);
+	for (i = 0; i < page_count; ++i)
+		result [i] = g_strdup_printf ("%d", i+1);
+
+	return result;	
+}
+
 
 #define MAGSTEP  1.2
 #define IMAGSTEP 0.8333333333333
@@ -381,6 +697,7 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 
 	gpdf_view = GPDF_VIEW (widget);
 	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view), &x, &y);
+	do_scroll = FALSE;
 	height = widget->allocation.height;
 	width = widget->allocation.width;
 	big_step = event->state & GDK_SHIFT_MASK;
@@ -414,8 +731,9 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Up:
 	case GDK_Up:
+		do_scroll = TRUE;
 		if (big_step) {
-			goto page_up;
+			goto prev_screen;
 		} else {
 			do_scroll = TRUE;
 			y -= 10;
@@ -423,8 +741,9 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Down:
 	case GDK_Down:
+		do_scroll = TRUE;
 		if (big_step) {
-			goto page_down;
+			goto next_screen;
 		} else {
 			do_scroll = TRUE;
 			y += 10;
@@ -432,10 +751,13 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Page_Up:
 	case GDK_Page_Up:
+		if (gpdf_view->priv->current_page > 1)
+			gpdf_view_page_prev (gpdf_view);
+		break;
 	case GDK_Delete:
 	case GDK_KP_Delete:
 	case GDK_BackSpace:
-	page_up:
+	prev_screen:
 		do_scroll = TRUE;
 		if (y <= 0) {
 			if (gpdf_view->priv->current_page > 1) {
@@ -452,9 +774,14 @@ gpdf_view_key_press_event (GtkWidget *widget, GdkEventKey *event)
 		break;
 	case GDK_KP_Page_Down:
 	case GDK_Page_Down:
+		if (gpdf_view->priv->current_page < 
+		    gpdf_view->priv->pdf_doc->getNumPages ())
+			gpdf_view_page_next (gpdf_view);
+		break;
 	case GDK_space:
-	page_down:
-		if (y >= GTK_LAYOUT (widget)->height - height) {
+	next_screen:
+		do_scroll = TRUE;
+		if (y >= (gint)(GTK_LAYOUT (widget)->height) - height) {
 			if (gpdf_view->priv->current_page < 
 			    gpdf_view->priv->pdf_doc->getNumPages ()) {
 				gpdf_view_page_next (gpdf_view);
@@ -483,6 +810,12 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 
 	priv = GPDF_VIEW (widget)->priv;
 
+	if (GNOME_CALL_PARENT_WITH_DEFAULT (GTK_WIDGET_CLASS,
+					    button_press_event,
+					    (widget, event), FALSE) == TRUE) {
+		return TRUE;
+	}
+
 	if (!GTK_WIDGET_HAS_FOCUS (widget))
 		gtk_widget_grab_focus (widget);
 
@@ -504,7 +837,7 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 
 		cursor = gdk_cursor_new (GDK_FLEUR);
 		gdk_pointer_grab (widget->window, FALSE,
-				  (enum GdkEventMask)
+				  (GdkEventMask)
 				  (GDK_POINTER_MOTION_MASK | 
 				   GDK_POINTER_MOTION_HINT_MASK |
 				   GDK_BUTTON_RELEASE_MASK),
@@ -512,10 +845,8 @@ gpdf_view_button_press_event (GtkWidget *widget, GdkEventButton *event)
 		gdk_cursor_unref (cursor);
 		return TRUE;
 	default:
-		break;
+		return FALSE;
 	}
-
-	return FALSE;
 }
 
 static gboolean
@@ -536,22 +867,20 @@ gpdf_view_button_release_event (GtkWidget *widget, GdkEventButton *event)
 	return TRUE;
 }
 
-static gboolean
-gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
+static void
+gpdf_view_handle_drag_motion (GPdfView *gpdf_view, GdkEventMotion *event)
 {
 	GPdfViewPrivate *priv;
 	gint x, y, dx, dy;
 	GdkModifierType mod;
 
-	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
-	
-	priv = GPDF_VIEW (widget)->priv;
+	g_return_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view));
 
-	if (!priv->dragging)
-		return FALSE;
+	priv = gpdf_view->priv;
 
 	if (event->is_hint) {
-		gdk_window_get_pointer (widget->window, &x, &y, &mod);
+		gdk_window_get_pointer (GTK_WIDGET (gpdf_view)->window,
+					&x, &y, &mod);
 	} else {
 		x = (gint)event->x;
 		y = (gint)event->y;
@@ -560,17 +889,28 @@ gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
 	dx = priv->drag_anchor_x - x;
 	dy = priv->drag_anchor_y - y;
 
-	gnome_canvas_scroll_to (GNOME_CANVAS (widget), 
+	gnome_canvas_scroll_to (GNOME_CANVAS (gpdf_view), 
 				priv->drag_offset_x + dx,
 				priv->drag_offset_y + dy);
 
 	priv->drag_anchor_x = (gint)event->x;
 	priv->drag_anchor_y = (gint)event->y;
-	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (widget),
+	gnome_canvas_get_scroll_offsets (GNOME_CANVAS (gpdf_view),
 					 &priv->drag_offset_x,
 					 &priv->drag_offset_y);
+}
 
-	return TRUE;
+static gboolean
+gpdf_view_motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
+{
+
+	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (widget), FALSE);
+	
+	if (GPDF_VIEW (widget)->priv->dragging)
+		gpdf_view_handle_drag_motion (GPDF_VIEW (widget), event);
+
+	return GNOME_CALL_PARENT_WITH_DEFAULT (
+		GTK_WIDGET_CLASS, motion_notify_event, (widget, event), TRUE);
 }
 
 static void
@@ -634,7 +974,8 @@ gpdf_view_set_pdf_doc (GPdfView *gpdf_view, PDFDoc *pdf_doc)
 	priv->current_page = 0;
 	priv->output_dev->startDoc (pdf_doc->getXRef());
 	gpdf_view_page_background_show (gpdf_view);
-	gpdf_view_goto_page (gpdf_view, 1);
+	/* FIXME clear history */
+	gpdf_view_goto_page_no_history (gpdf_view, 1);
 }
 
 
@@ -657,6 +998,11 @@ gpdf_view_dispose (GObject *object)
 		priv->print_context = NULL;
 	}
 
+	if (priv->links_layer) {
+		g_object_unref (priv->links_layer);
+		priv->links_layer = NULL;
+	}
+
 	if (priv->gesture_handler) {
 		gtk_gesture_handler_destroy (priv->gesture_handler);
 		priv->gesture_handler = NULL;
@@ -713,6 +1059,31 @@ gpdf_view_class_init (GPdfViewClass *klass)
 		NULL, NULL,
 		gpdf_marshal_VOID__INT,
 		G_TYPE_NONE, 1, G_TYPE_INT);
+
+	gpdf_view_signals [CLOSE_REQUESTED_SIGNAL] = g_signal_new (
+		"close_requested",
+		G_TYPE_FROM_CLASS (object_class),
+		G_SIGNAL_RUN_LAST,
+		G_STRUCT_OFFSET (GPdfViewClass, close_requested),
+		NULL, NULL,
+		gpdf_marshal_VOID__VOID,
+		G_TYPE_NONE, 0);
+
+	gpdf_view_signals [QUIT_REQUESTED_SIGNAL] = g_signal_new (
+		"quit_requested",
+		G_TYPE_FROM_CLASS (object_class),
+		G_SIGNAL_RUN_LAST,
+		G_STRUCT_OFFSET (GPdfViewClass, quit_requested),
+		NULL, NULL,
+		gpdf_marshal_VOID__VOID,
+		G_TYPE_NONE, 0);
+}
+
+static void
+ggv_document_interface_init (GgvDocumentClass *klass, gpointer iface_data)
+{
+	klass->get_page_count = gpdf_view_get_page_count;
+	klass->get_page_names = gpdf_view_get_page_names;
 }
 
 static void
@@ -793,8 +1164,6 @@ GPdfView *
 gpdf_view_construct (GPdfView *gpdf_view)
 {
 	GPdfViewPrivate *priv;
-	GnomePrintConfig *config;
-	GnomePrintContext *pc;
 
 	g_return_val_if_fail (GPDF_IS_NON_NULL_VIEW (gpdf_view), NULL);
 
@@ -820,4 +1189,44 @@ gpdf_view_new (void)
 	return GTK_WIDGET (gpdf_view_construct (gpdf_view));
 }
 
+static void
+gpdf_view_class_init_trampoline (gpointer klass, gpointer data)
+{
+	parent_class = (GnomeCanvasClass *)g_type_class_ref (PARENT_TYPE);
+	gpdf_view_class_init ((GPdfViewClass *)klass);
+}
+GType
+gpdf_view_get_type (void)
+{
+	static GType object_type = 0;
+	if (object_type == 0) {
+		static const GTypeInfo object_info = {
+		    sizeof (GPdfViewClass),
+		    NULL,		/* base_init */
+		    NULL,		/* base_finalize */
+		    gpdf_view_class_init_trampoline,
+		    NULL,		/* class_finalize */
+		    NULL,               /* class_data */
+ 		    sizeof (GPdfView),
+		    0,                  /* n_preallocs */
+		    (GInstanceInitFunc) gpdf_view_instance_init
+		};
+
+		static const GInterfaceInfo ggv_document_info = {
+			(GInterfaceInitFunc) ggv_document_interface_init,
+			NULL,
+			NULL
+		};
+
+		object_type = g_type_register_static (PARENT_TYPE, "GPdfView",
+						      &object_info,
+						      (GTypeFlags)0);
+
+		g_type_add_interface_static (object_type,
+					     GGV_DOCUMENT_TYPE,
+					     &ggv_document_info);
+	}
+	return object_type;
+}
+
 END_EXTERN_C
